"""CLI functions for Keyring TOTP Generator."""
# ****************************************************************************
# *   Keyring TOTP Generator                                                 *
# *                                                                          *
# *   Copyright (C) 2017 by Jeremy Falling except where noted.               *
# *                                                                          *
# *   This program is free software: you can redistribute it and/or modify   *
# *   it under the terms of the GNU General Public License as published by   *
# *   the Free Software Foundation, either version 3 of the License, or      *
# *   (at your option) any later version.                                    *
# *                                                                          *
# *   This program is distributed in the hope that it will be useful,        *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
# *   GNU General Public License for more details.                           *
# *                                                                          *
# *   You should have received a copy of the GNU General Public License      *
# *   along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
# ****************************************************************************

import argparse
import binascii
import logging
import signal
import sys

import keyring
import pyperclip

import totp_generator
from totp_generator.core_utils import KeyringTotpGenerator

# Make this optional since installing it may require elevated privileges
try:
    from setproctitle import setproctitle
except ImportError:  # pragma: no cover
    pass

# load logger before keyring to stop log notice on some platforms
logging.basicConfig(level=logging.WARN)
logging.getLogger()
logger = logging.getLogger()

PROGNAME = totp_generator.__progname__
VERSION = totp_generator.__version__
SERVICE_NAME = 'totp_generator'

YES_ANSWERS = ['y', 'yes']

# set the process name before going any further. This allows keychain requests to show as this program instead
# of simply 'python'. Also replace the spaces to work better with some systems.
if 'setproctitle' in sys.modules:
    setproctitle(PROGNAME.replace(' ', '-'))

# backwards compatibility for py2
try:
    input = raw_input
except NameError:
    pass


def signal_handler(signal, frame):  # pragma: no cover
    """Catch interrupts and exit without a stack trace."""
    print('\nExiting...\n')
    sys.exit(0)


def version_string():
    """Show version info and exit."""
    return '{name} version {ver}\n'.format(name=PROGNAME, ver=VERSION)


def service_menu(services, list_services=False):
    """
    Interactive service selection.

    :param services: TOTP services data
    :param list_services: only list services without number and return None. Defaults to False
    :return: totp service name or None
    """
    i = 0
    options = list()

    if len(services) == 0:
        print(
            'It appears you have not loaded any TOTP data. Please add a TOTP service. Run with --help for more.\n')
        sys.exit(1)

    for service in services:
        i += 1
        options.append(service)

        if list_services:
            # don't print service number when prompt is disabled
            print(service)
        else:
            print('{i}: {name}'.format(i=i, name=service))

    if list_services:
        return

    while True:
        user_in = input("\nSelect a service by number: ")
        try:
            sel = int(user_in) - 1
            # range is exclusive
            if int(sel) in range(0, i):
                break
        except ValueError:
            pass

        print("Your selection is not valid. Try again.")

    return options[int(sel)]


def main():
    """Main cli function."""
    # catch ctrl+c
    signal.signal(signal.SIGINT, signal_handler)

    parser = argparse.ArgumentParser(description=PROGNAME +
                                                 '\n\nUtility that generates TOTP codes and stores the TOTP secrets ' +
                                                 'in your system keyring.\nTOTP' +
                                                 'Secrets are stored in a keyring supported by the keyring module.\n',
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-a', '--add', action='store_true', help='add a TOTP service')
    parser.add_argument('-c', '--copy', action='store_true', help='copy TOTP code to clipboard after generating')
    parser.add_argument('-d', '--debug', action='store_true', help='enable debug logging')
    parser.add_argument('-e', '--edit', action='store_true', help='edit a TOTP service')
    parser.add_argument('--export', dest='export_file', action='store',
                        help='export all credentials to a plain text json file')
    parser.add_argument('--import', dest='import_file', action='store', help='import JSON dump of credentials')
    parser.add_argument('-l', '--list', action='store_true', help='list TOTP services')
    parser.add_argument('-r', '--remove', action='store_true', help='remove a TOTP service')
    parser.add_argument('-s', '--service', type=str, default=None, help='specify a TOTP service instead of picking ' +
                                                                        'from a list.')
    parser.add_argument('-v', '--version', action='store_true', help='show version and exit')
    parser.add_argument('-q', '--quiet', action='store_true', help='does nothing unless combined with another flag:\n' +
                                                                   'when used with the copy flag no TOTP code is ' +
                                                                   'not shown but all other output is shown.\n' +
                                                                   'when used with the service flag no output other ' +
                                                                   'than the TOTP code is shown.\n' +
                                                                   'when used with the copy and service flags no ' +
                                                                   'output is shown at all.\n' +
                                                                   'does not apply to import, export, add, edit, ' +
                                                                   'remove, debug, and help flags.\n' +
                                                                   'errors will always be shown.')
    args = parser.parse_args()

    keyring_generator = KeyringTotpGenerator()

    # handle flags
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    if 'setproctitle' not in sys.modules:
        logger.info('setproctitle Module is not loaded. Unable to set process title.')
    logger.debug('keyring module config root: ' + keyring.util.platform_.config_root())
    logger.debug('keyring that will be used: ' + keyring.get_keyring().name)

    if args.version:
        print(version_string())
        sys.exit(0)

    if args.add:
        name = input("Enter the name of this service: ")
        code = input("Enter the TOTP secret: ")
        res = keyring_generator.add_service(name, code)
        if res:
            print("TOTP service added")
        else:
            print("There was an error adding this TOTP service. See above for a detailed error.")
        sys.exit(0)

    if args.remove:
        service = service_menu(keyring_generator.get_services())
        res = keyring_generator.rm_service(service)
        if res:
            print('Removed service {s}\n'.format(s=service))
        else:
            print('Error removing service {s}. See above for more details\n'.format(s=service))
        sys.exit(0)

    if args.edit:
        service = service_menu(keyring_generator.get_services())

        while True:
            name = input("Enter the new name of this service or hit return to leave unchanged: ")
            if not name:
                name = None

            if name not in keyring_generator.get_services():
                break

            print('That service name already exists. Please select a new name!\n')

        code = input("Enter the new TOTP secret or hit return to leave unchanged: ")
        if not code:
            code = None

        ret = keyring_generator.edit_service(service, name, code)
        if ret:
            print("Service {s} updated.".format(s=service))
        else:
            print("Error updating service {s}. See above for a detailed error.".format(s=service))

        sys.exit(0)

    if args.import_file:
        if input("Warning: You are about to import all TOTP credentials from {f}. \nExisting entries that have the\
 same name as imported entries will be overwritten without warning.\nDo you want to continue? [y/n]: ".
                         format(f=args.import_file)).lower() not in YES_ANSWERS:
            print("Not performing import.\n")
            sys.exit(1)

        import_res = keyring_generator.import_creds_from_file(args.import_file)
        if import_res:
            print("Successfully imported credentials.\n")
        sys.exit(0)

    if args.export_file:
        if input("Warning: You are about to export all TOTP credentials in PLAIN TEXT from your current keying to {f}. \
\nDo you want to continue? [y/n]: ".format(f=args.export_file)).lower() not in YES_ANSWERS:
            print("Not performing export.\n")
            sys.exit(1)

        keyring_generator.export_creds_to_file(args.export_file)
        print("Successfully exported credentials to {n}\n".format(n=args.export_file))
        sys.exit(0)

    if args.list:
        service_menu(keyring_generator.get_services(), list_services=True)
        sys.exit(0)


    if not args.service:
        service = service_menu(keyring_generator.get_services())

    else:
        service = args.service
        if service not in keyring_generator.get_services():
            print('That service does not exist\n')
            sys.exit(1)

    try:
        totp_code = keyring_generator.get_totp_code(service)
        if args.quiet and args.copy:
            pass
        else:
            print(totp_code)

            if not args.service:
                print('')

        if args.copy:
            pyperclip.copy(totp_code)

    except TypeError as e:
        print("Error generating TOTP code: {e}\n".format(e=e))
        exit(1)

    except binascii.Error as e:
        if 'Non-base32 digit found' in str(e):
            print("Error generating TOTP code: 'Non-base32 digit found'. Usually this means you did not provide a valid \
TOTP code for the service {s}\n".format(s=service))
        elif 'Incorrect padding' in str(e):
            print("Error generating TOTP code: 'Incorrect padding'. Usually this means you did not provide a valid \
TOTP code for the service {s}\n".format(s=service))
        else:
            print("Error generating TOTP code: {e}\n".format(e=e))
        exit(1)


if __name__ == '__main__':
    main()



"""Core functions for Keyring TOTP Generator."""
# ****************************************************************************
# *   Keyring TOTP Generator                                                 *
# *                                                                          *
# *   Copyright (C) 2017 by Jeremy Falling except where noted.               *
# *                                                                          *
# *   This program is free software: you can redistribute it and/or modify   *
# *   it under the terms of the GNU General Public License as published by   *
# *   the Free Software Foundation, either version 3 of the License, or      *
# *   (at your option) any later version.                                    *
# *                                                                          *
# *   This program is distributed in the hope that it will be useful,        *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
# *   GNU General Public License for more details.                           *
# *                                                                          *
# *   You should have received a copy of the GNU General Public License      *
# *   along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
# ****************************************************************************

import codecs
import json
import logging
from getpass import getuser

import keyring
import onetimepass

import totp_generator

# load logger before keyring to stop log notice on some platforms
logging.basicConfig(level=logging.WARN)
logging.getLogger()
logger = logging.getLogger()

PROGNAME = totp_generator.__progname__
VERSION = totp_generator.__version__
SERVICE_NAME = 'totp_generator'


class KeyringTotpGenerator:

    def __init__(self, force_keyring=None):
        """Return KeyringTotpGenerator object."""
        self.user = getuser()
        if force_keyring:
            keyring.set_keyring(force_keyring)
        self.creds = self.load_creds()

    def load_creds(self, init_new=True):
        """Load TOTP credentials from keyring."""
        keyring_data = keyring.get_password(SERVICE_NAME, self.user)
        if not keyring_data:
            if init_new:
                # keyring does not exist, init a new one.
                return dict()
            else:
                return False

        # data found, parse json
        json_data = json.loads(keyring_data)
        return json_data

    def add_service(self, name, code):
        if name in self.creds:
            logger.error(
                'There is already a service with the name {n}. Use the edit option to update this service.'.format(
                    n=name))
            return False

        self.creds[name] = dict()
        self.creds[name]['code'] = code

        self.save_creds()
        return True

    def edit_service(self, old_name, new_name=None, code=None):
        if new_name in self.creds:
            logger.error(
                'There is already a service with the name {n}.'.format(
                    n=new_name))
            return False

        if not code and not new_name:
            logger.warning('You provided no changes. Leaving service as-is.\n')
            return False

        if code:
            self.creds[old_name]['code'] = code

        if new_name:
            existing_data = self.creds[old_name]
            self.creds.pop(old_name, None)
            self.creds[new_name] = existing_data

        self.save_creds()
        return True

    def export_creds_to_file(self, file_name):
        with open(file_name, 'wb') as f:
            json.dump(self.creds, codecs.getwriter('utf-8')(f), ensure_ascii=False)
        return

    def get_services(self):
        services = []
        for key in sorted(self.creds):
            services.append(key)
        return services

    def get_totp_code(self, service):
        # force six digits
        code = "%06d" % (onetimepass.get_totp(self.creds[service]['code']))
        return code

    def import_creds_from_file(self, file_name):
        try:
            file = open(file_name, 'r')
        except EnvironmentError as e:
            logger.fatal("Error: opening dump file {n}: {e}".format(n=file_name, e=e))
            return False
        try:
            loaded_config = json.loads(file.read())
        except ValueError as err:
            logger.fatal("Error: could not parse dump file. Ensure it is valid json: {err}.".format(err=err))
            return False

        self.creds.update(loaded_config)
        self.save_creds()
        return True

    def rm_service(self, service):
        self.creds.pop(service, None)
        self.save_creds()
        return True

    def save_creds(self):
        keyring.set_password(SERVICE_NAME, self.user, json.dumps(self.creds))
        return



import codecs
import os
import re
import sys
from shutil import rmtree

from setuptools import Command, setup

import totp_generator

AUTHOR = 'Jeremy Falling'
DESCRIPTION = 'Utility that generates TOTP codes and stores the TOTP secrets in your system keyring.'
URL = 'https://github.com/jjfalling/totp-generator'

# What packages are required for this module to be executed?
REQUIRED = [
    'keyring>=12.2.0,<18.0.0',
    'keyrings.alt>=3.1,<4.0',
    'onetimepass>=1.0.1,<1.1.0',
    'pyperclip>=1.6.1<1.7.0',
]

TESTS_REQUIRED = ["pytest", "pytest-cov<2.6", "mock; python_version < '3.4'", "keyrings.alt", "wheel"]

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r") as fh:
    long_description = fh.read()


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds...')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution...')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine...')
        os.system('twine upload dist/*')

        self.status('Pushing git tags...')
        os.system('git tag {0}'.format(totp_generator.__version__))
        os.system('git push --tags')

        sys.exit()


setup(
    name="TOTP Generator",
    version=totp_generator.__version__,
    author=AUTHOR,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    license='GPLv3',
    install_requires=REQUIRED,
    include_package_data=True,
    packages=['totp_generator'],
    entry_points={
        "console_scripts": [
            "totp_generator = totp_generator.cli:main"
        ]
    },
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
    ],
    tests_require=TESTS_REQUIRED,
    extras_require={
        'dev': ['setuptools', 'twine', 'wheel'],
        'test': TESTS_REQUIRED,
        'proctitle': 'setproctitle>=1.1.10,<1.2.0'
    },
    # setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)
